#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/module.h>

#include <linux/kernel.h> 
#include <linux/malloc.h> 

#include <linux/input.h>
#include <linux/usb.h>

#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <asm/uaccess.h>

#include <linux/usbdevice_fs.h>

#include <linux/config.h>
#include <linux/string.h>
#include <linux/bitops.h>
#include <linux/slab.h>
#include <linux/interrupt.h>  
#include <linux/kmod.h>
#include <linux/init.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/spinlock.h>

#include <linux/smp_lock.h>


/*****************************************************************************
*
*****************************************************************************/
static struct usb_device_id reader_device_ids [] = {
    /* ASE III */
    { USB_DEVICE(0x0DC3, 0x0801) },	
    { }				
};

MODULE_DEVICE_TABLE (usb, reader_device_ids);


/*****************************************************************************
*
*****************************************************************************/
static struct usb_driver reader_driver; 


/*****************************************************************************
*
*****************************************************************************/
#define IS_EP_BULK(ep)  ((ep).bmAttributes == USB_ENDPOINT_XFER_BULK ? 1 : 0)
#define IS_EP_BULK_IN(ep) (IS_EP_BULK(ep) && ((ep).bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN)
#define IS_EP_BULK_OUT(ep) (IS_EP_BULK(ep) && ((ep).bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_OUT)
#define IS_EP_INTR(ep) ((ep).bmAttributes == USB_ENDPOINT_XFER_INT ? 1 : 0)


#define OBUF_SIZE 32 

#define EVENT_BUFFER_SIZE           8
#define RESPONSE_BUFFER_SIZE        32
#define BULK_BUFFER_SIZE            300

#define RESPONSE_BACK_BUFFER_SIZE   8192 /*2048*/
#define EVENT_BACK_BUFFER_SIZE      2 * EVENT_BUFFER_SIZE

/*****************************************************************************
* IOCTL stuff
*****************************************************************************/
#define IOCTL_RESPONSE_BUFFER_SIZE  512
#define IOCTL_EVENT_BUFFER_SIZE     16

typedef struct {
    int     len;
    int     timeout; /* in ms */
    char    data[IOCTL_RESPONSE_BUFFER_SIZE];
} ioctlResponseStruct;


typedef struct {
    int     timeout; /* in ms */
    char    data[IOCTL_EVENT_BUFFER_SIZE];
} ioctlEventStruct;


typedef struct {
    char    present;
} ioctlStatusStruct;


typedef struct {
    char    clear; // redundant
} ioctlClearStruct;


#define IOCTL_ASE_RESPONSE      _IOWR('U', 0x20, ioctlResponseStruct)
#define IOCTL_ASE_EVENT         _IOWR('U', 0x21, ioctlEventStruct)
#define IOCTL_ASE_STATUS        _IOWR('U', 0x22, ioctlStatusStruct)
#define IOCTL_ASE_CLEAR_BUFFER  _IOWR('U', 0x23, ioctlClearStruct)


/*****************************************************************************
*
*****************************************************************************/
struct reader_data {
    struct usb_device*  reader_dev;
    devfs_handle_t      devfs;

    struct urb          response_irq;
    struct urb          event_irq;
    unsigned int        ifnum;	        

    int                 minor;

    unsigned char       eventBuffer[EVENT_BUFFER_SIZE]; 
    unsigned char       responseBuffer[RESPONSE_BUFFER_SIZE]; 
    unsigned char       bulkBuffer[BULK_BUFFER_SIZE]; 

    /* cyclic buffers */
    unsigned char       responseBackBuffer[RESPONSE_BACK_BUFFER_SIZE];
    unsigned char       eventBackBuffer[EVENT_BACK_BUFFER_SIZE];

    /* cur points to the next unread byte; if cur != last, some data hasnt been read yet */
    int                 curResponsePos, lastResponsePos;
    int                 curEventPos, lastEventPos;

    rwlock_t            rwlockResponse;
    rwlock_t            rwlockEvent;

    unsigned char       isopen;		    /* Nonzero if the device is open  */
    char                present;        /* Nonzero if device is present */
    char                bulk_out_ep;
    char                response_intr_ep;
    char                event_intr_ep;  /* Endpoint assignments */
    
    struct semaphore    gen_lock;       /* lock to prevent concurrent reads or writes */
};

/*****************************************************************************
*
*****************************************************************************/
#define READER_MAJOR        180	
#define READER_BASE_MNR     224
	
#define READER_MAX_MINOR    16

#define USB_READER_MINOR(X) MINOR((X)->i_rdev) - READER_BASE_MNR

static struct reader_data* reader_data_table[READER_MAX_MINOR] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
                                                                  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};

static DECLARE_MUTEX (rdr_mutex); /* initialized to be unlocked */

extern devfs_handle_t usb_devfs_handle;


/*****************************************************************************
*
*****************************************************************************/
static int open_reader(struct inode * inode, struct file * file)
{
    struct reader_data* data;
    struct usb_device*  dev;
    int err = 0;
	int minor;

	MOD_INC_USE_COUNT;

	down(&rdr_mutex);

	minor = USB_READER_MINOR(inode);

    dbg("\n\nopen_reader.");
    
	if (!reader_data_table[minor]) {
		up(&rdr_mutex);
		MOD_DEC_USE_COUNT;
		err("open_reader(%d): Unable to access minor data", minor);
		return -ENODEV;
	}

    data = reader_data_table[minor];
    
    dev = data->reader_dev;
    
	down(&(data->gen_lock));	

	up(&rdr_mutex); /* now handled by the above  */

    if (!dev) {
    	err("open_reader: Reader device not present");
	    err = -ENODEV;
        goto out_error;
    }
    
    if (!data->present) {
    	err("open_reader: Reader is not present");
	    err = -ENODEV;
        goto out_error;
    }

    data->rwlockEvent = RW_LOCK_UNLOCKED;
    data->rwlockResponse = RW_LOCK_UNLOCKED;

    data->curResponsePos = data->lastResponsePos = data->curEventPos = data->lastEventPos = 0;
    
    /* allow multiple opens */
    data->isopen++; 
    
    file->private_data = data; /* Used by the read and write metheds */
    
out_error:

	up(&(data->gen_lock)); /* Wake up any possible contending processes */

	if (err)
		MOD_DEC_USE_COUNT;

	return err;
}



/*****************************************************************************
*
*****************************************************************************/
static int close_reader(struct inode * inode, struct file * file)
{
    struct reader_data *data;
    
    dbg("close_reader.\n\n");
    
    down(&rdr_mutex);

    data = file->private_data;

    down(&(data->gen_lock));	

    data->isopen--;
    file->private_data = NULL;

    /* if the device has already been disconnected, we need to perform the
       cleanup code of disconnect() here! (the user should call close() when
       he finds out that the device has been removed) */
    if (!(data->present) && !(data->isopen)) {
        if(data->response_intr_ep) {
	        usb_unlink_urb(&data->response_irq); 
        }
        if(data->event_intr_ep) {
	        usb_unlink_urb(&data->event_irq); 
        }

        usb_driver_release_interface(&reader_driver, 
                                     &data->reader_dev->actconfig->interface[data->ifnum]);

        devfs_unregister(data->devfs);
        reader_data_table[data->minor] = NULL;
        up(&(data->gen_lock));
        kfree (data);
        up(&rdr_mutex);
    }
    else {
        up(&rdr_mutex);
	    up(&(data->gen_lock)); 
    }

    MOD_DEC_USE_COUNT;
    
    return 0;
}


/*****************************************************************************
*
*****************************************************************************/
static ssize_t write_reader    (struct file * file, const char * buffer,
                                size_t count, loff_t *ppos)
{
    struct reader_data* data;
    struct usb_device*  dev;
    
    ssize_t bytes_written = 0; /* Overall count of bytes written */
    ssize_t ret = 0;
    
    int this_write;		/* Number of bytes to write */
    int actualWrittenLen;		/* Number of bytes successfully written */
    int result = 0;
    
    char *obuf;
    int i;
    
    dbg("Athena write_reader: trying to write %d bytes", count);

    data = file->private_data;
    
    down(&(data->gen_lock));

    if (!(data->present)) {
        up(&(data->gen_lock));
        return -ENODEV;
    }

    obuf = data->bulkBuffer;
    
    dev = data->reader_dev;

    file->f_dentry->d_inode->i_atime = CURRENT_TIME;
    
    while (count > 0) {
	    if (signal_pending(current)) {
	        ret = -EINTR;
	        break;
	    }
	    
	    this_write = (count >= OBUF_SIZE) ? OBUF_SIZE : count;
	    
	    if (copy_from_user(data->bulkBuffer, buffer, this_write)) {
	        ret = -EFAULT;
	        break;
	    }

#ifdef ASE_DEBUG
	    dbg("chunk");
        for (i = 0 ; i < this_write ; ++i)
	        dbg("0x%x", data->bulkBuffer[i]);
#endif
    
        if (!(data->present)) {
            up(&(data->gen_lock));
            return -ENODEV;
        }

        result = usb_bulk_msg(dev,usb_sndbulkpipe(dev, data->bulk_out_ep), obuf, this_write, &actualWrittenLen, 60*HZ);
	    /*	dbg("write_reader stats: result:%d this_write:%d actualWrittenLen:%d", result, this_write, actualWrittenLen); */
	    
	    if (result == -ETIMEDOUT) {	/* NAK -- shouldn't happen */
	        dbg("write_reader: NAK recieved.");
            ret = result;
			    break;
	    } 
        else if (result < 0) { /* We should not get any I/O errors */
	        dbg("write_reader: funky result: %d. Please notify the maintainer.", result);
            ret = -EIO;
	        break;
	    }
	    
	    if (actualWrittenLen != this_write) { /* Unable to write all contents of obuf */
	        ret = -EIO;
	        break;
	    }
	    
	    if (actualWrittenLen) { /* Data written */
	        buffer += actualWrittenLen;
	        count -= actualWrittenLen;
	        bytes_written += actualWrittenLen;
	    } 
        else { /* No data written */
	        ret = 0;
	        break;
	    }
    }
    
    up(&(data->gen_lock));
    return ret ? ret : bytes_written;
}


/*****************************************************************************
*
*****************************************************************************/
/* forward declarations */
static void event_irq_callback(struct urb *urb);
static void response_irq_callback(struct urb *urb);


/*****************************************************************************
*
*****************************************************************************/
/* Returns the number of bytes copied to arg.
 NOTE: arg should be large enough to contain all the data to be returned.*/
static int ioctl_reader (struct inode *inode, struct file *file, 
                         unsigned int cmd, unsigned long arg)
{
    struct reader_data* data = file->private_data;
    struct usb_device* dev = data->reader_dev;
    int ret = 0, count = 0;
    unsigned long flags1, flags2;
    int len, nextAmount, i;

    switch (cmd) {
        case IOCTL_ASE_RESPONSE: {
            ioctlResponseStruct resp; 
	    
            if (!(data->present)) {
                return -ENODEV;
            }

            if (copy_from_user(&resp, (void *)arg, sizeof(resp))) {
        		return -EFAULT;
    	    }
            
            len = resp.len;
	    
            do { /* loop until something arrives or timeout */
                read_lock_irqsave(&(data->rwlockResponse), flags2);
		
                if (data->curResponsePos == data->lastResponsePos) { /* no unread data */
                    read_unlock_irqrestore(&(data->rwlockResponse), flags2);
                    resp.timeout--;
                }
                else {
                    read_unlock_irqrestore(&(data->rwlockResponse), flags2);
		    
                    /* now we need to update the curPos, so we need to write lock it*/
                    write_lock_irqsave(&(data->rwlockResponse), flags2);
		            /*		    
		            dbg("Athena reader_ioctl: curPos = %d, lastPos = %d, len = %d", data->curResponsePos, 
                         data->lastResponsePos, len);
		            */
                    if (data->curResponsePos < data->lastResponsePos) {
                        count = (data->lastResponsePos - data->curResponsePos) > len ? 
                			    len : (data->lastResponsePos - data->curResponsePos);
                        memcpy((char*)(resp.data), &(data->responseBackBuffer[data->curResponsePos]), count);
            			ret += count;
                        data->curResponsePos += count;
                    }
                    else {
                        count = (RESPONSE_BACK_BUFFER_SIZE - data->curResponsePos) > len ? 
                			    len : (RESPONSE_BACK_BUFFER_SIZE - data->curResponsePos);
                        memcpy((char*)(resp.data), &(data->responseBackBuffer[data->curResponsePos]), count);
			            data->curResponsePos += count;
			            ret += count;
                        if (ret < len) { /* more bytes to copy (len-ret at most) */
                            count = (data->lastResponsePos < (len - ret) ? 
                        			 data->lastResponsePos : len - ret);
                            if (count) {
                                memcpy((char*)(&(resp.data[ret])), data->responseBackBuffer, count);
                                ret += count;
                            }
                            data->curResponsePos = count;
                        }
                    }
		    
                    write_unlock_irqrestore(&(data->rwlockResponse), flags2);
                }
		
            } while (ret != len && resp.timeout);
	    
#ifdef ASE_DEBUG
	    dbg("IOCTL_ASE_RESPONSE: len = %d, data: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x...", 
		 ret, (unsigned char)resp.data[0], (unsigned char)resp.data[1], (unsigned char)resp.data[2], 
		 (unsigned char)resp.data[3], (unsigned char)resp.data[4], (unsigned char)resp.data[5], 
		 (unsigned char)resp.data[6], (unsigned char)resp.data[7]);
#endif
	        if (copy_to_user((void*)arg, &resp, sizeof(resp))) {
    		    ret = -EFAULT;
	        }
	    
            break;
            }

        case IOCTL_ASE_EVENT: {
            ioctlEventStruct resp; 

            if (!(data->present)) {
                return -ENODEV;
            }

            if (copy_from_user(&resp, (void *)arg, sizeof(resp))) {
                return -EFAULT;
            }

            len = EVENT_BACK_BUFFER_SIZE; /* return all the events in back buffer */

            do { /* loop until something arrives or timeout */
                read_lock_irqsave(&(data->rwlockEvent), flags1);

                if (data->curEventPos == data->lastEventPos) { /* no unread data */
                    read_unlock_irqrestore(&(data->rwlockEvent), flags1);
                    resp.timeout--;
                }
                else {
                    read_unlock_irqrestore(&(data->rwlockEvent), flags1);
		    
		            /* now we need to update the curPos, so we need to write lock it */
    	            write_lock_irqsave(&(data->rwlockEvent), flags1);
		            if (data->curEventPos < data->lastEventPos) {
        			    count = (data->lastEventPos - data->curEventPos) > len ? 
                                len : (data->lastEventPos - data->curEventPos);
			            memcpy((char*)(resp.data), &(data->eventBackBuffer[data->curEventPos]), count);
			            ret += count;
			            data->curEventPos += count;
	    	        }
    		        else {
		        	    count = (EVENT_BACK_BUFFER_SIZE - data->curEventPos) > len ? 
                                len : (EVENT_BACK_BUFFER_SIZE - data->curEventPos);
			            memcpy((char*)(resp.data), &(data->eventBackBuffer[data->curEventPos]), count);
			            data->curEventPos += count;
			            ret += count;
			            if (ret < len) { /* more bytes to copy (len-ret at most) */
			                count = (data->lastEventPos < (len - ret) ? 
    				                 data->lastEventPos : len - ret);
			                if (count) {
				                memcpy((char*)(&(resp.data[ret])), data->eventBackBuffer, count);
				                ret += count;
    			            }
	    	    	        data->curEventPos = count;
		    	        }
		            }
		        
		            write_unlock_irqrestore(&(data->rwlockEvent), flags1);
    		    }
		
            } while (ret != len && resp.timeout);

            if (copy_to_user((void*)arg, &resp, sizeof(resp))) {
                ret = -EFAULT;
            }

            break;

            }

        case IOCTL_ASE_STATUS: {
            ioctlStatusStruct resp; 

            resp.present = data->present;

            if (copy_to_user((void*)arg, &resp, sizeof(resp))) {
                ret = -EFAULT;
            }

	        break;
        }

        case IOCTL_ASE_CLEAR_BUFFER: {

            write_lock_irqsave(&(data->rwlockResponse), flags1);

            data->curResponsePos = data->lastResponsePos = 0;

            write_unlock_irqrestore(&(data->rwlockResponse), flags1);

    	    break;
        }

        default:
	        ret = -ENOIOCTLCMD;
	        break;
    }

    if (!(data->present)) {
#ifdef ASE_DEBUG
        info("##### ioctl_reader - no device is present!!!");
#endif
        return -ENODEV;
    }

    return ret;
}



/*****************************************************************************
*
*****************************************************************************/
struct file_operations usb_reader_fops = {
    write:		write_reader,
    ioctl:		ioctl_reader,
    open:		open_reader,
    release:	close_reader,
};



/*****************************************************************************
*
*****************************************************************************/
static void response_irq_callback (struct urb *urb)
{
    struct reader_data *data = urb->context;
    unsigned long flags;
    int i;
    unsigned char *theData = data->responseBuffer;
    
    /* if the status in nonzero (negative) than the request is either in progress
       or some error has occurred. */
    if (urb->status) {
	    return;
    }

    if (urb->actual_length > 0) {
        write_lock_irqsave(&(data->rwlockResponse), flags);

        for (i = 0 ; i < urb->actual_length ; ++i) {
            data->responseBackBuffer[data->lastResponsePos++] = data->responseBuffer[i];
            data->lastResponsePos %= RESPONSE_BACK_BUFFER_SIZE;
        }
        write_unlock_irqrestore(&(data->rwlockResponse), flags);
	
#ifdef ASE_DEBUG
	dbg("response_irq_callback: length = %d : 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x", 
	     urb->actual_length, theData[0], theData[1], theData[2], theData[3], 
	     theData[4], theData[5], theData[6], theData[7]);
	dbg("Athena: cont(1)        : 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x", 
	     theData[8], theData[9], theData[10], theData[11],
	     theData[12], theData[13], theData[14], theData[15]);
	dbg("Athena: cont(2)        : 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x", 
	     theData[16], theData[17], theData[18], theData[19], 
	     theData[20], theData[21], theData[22], theData[23]);
	dbg("Athena: cont(3)        : 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x", 
	     theData[24], theData[25], theData[26], theData[27], 
	     theData[28], theData[29], theData[30], theData[31]);
#endif
	
    }

    return;
}

/*****************************************************************************
*
*****************************************************************************/
static void event_irq_callback (struct urb *urb)
{
    struct reader_data *data = urb->context;
    unsigned long flags;
    int i;
    unsigned char *theData = data->eventBuffer;
    
    /* if the status in nonzero (negative) than the request is either in progress
       or some error has occurred. */
    if (urb->status) {
	    return;
    }
    
    if (urb->actual_length > 0) {
        write_lock_irqsave(&(data->rwlockEvent), flags);

        for (i = 0 ; i < urb->actual_length ; ++i) {
            data->eventBackBuffer[data->lastEventPos++] = data->eventBuffer[i];
            data->lastEventPos %= EVENT_BACK_BUFFER_SIZE;
        }
	    
        write_unlock_irqrestore(&(data->rwlockEvent), flags);

#ifdef ASE_DEBUG
	dbg("event_irq_callback: length = %d : 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x", 
	     urb->actual_length, theData[0], theData[1], theData[2],
	     theData[3], theData[4], theData[5], theData[6], theData[7]);
#endif

    }    

    return;
}

/*****************************************************************************
*
*****************************************************************************/
static void* probe_reader  (struct usb_device *dev, unsigned int ifnum,
	                        const struct usb_device_id *id)
{
    struct reader_data*                 data;
    struct usb_interface_descriptor*    interface;
    struct usb_endpoint_descriptor*     endpoint;
    
    int ep_cnt;
    int ix;
    
    int minor;
    char name[20];

    
    char valid_device = 0;
    char have_bulk_out, have_response_intr, have_event_intr;

    /*
     1. Check Vendor/Product
     2. Determine/Assign Bulk Endpoints
     3. Determine/Assign Intr Endpoint
    */
    if (dev->descriptor.idVendor == 0x0DC3 &&  
	    dev->descriptor.idProduct == 0x0801) { 
	    valid_device = 1;
    }
    
    if (!valid_device)
	    return NULL;   
    
    if (dev->descriptor.bNumConfigurations != 1) {
	    dbg("probe_reader: Only one device configuration is supported.");
	    return NULL;
    }
    
    if (dev->config[0].bNumInterfaces != 1) {
	    dbg("probe_reader: Only one device interface is supported.");
	    return NULL;
    }
    
    interface = dev->config[0].interface[ifnum].altsetting;
    endpoint = interface[ifnum].endpoint;
    
    /* Check for 1 bulk endpoint + 2 interrupt endpoints */
    
    if (interface->bNumEndpoints != 3) {
	    dbg("probe_reader: Only three endpoints supported.");
	    return NULL;
    }
    
    ep_cnt = have_bulk_out = have_response_intr = have_event_intr = 0;
    
    while (ep_cnt < interface->bNumEndpoints) {
	
        dbg("Endpoint %d: Length = %d", ep_cnt, endpoint[ep_cnt].bLength);
	    dbg("Endpoint %d: DescType = %d", ep_cnt, endpoint[ep_cnt].bDescriptorType);
	    dbg("Endpoint %d: EPAddress = %x", ep_cnt, endpoint[ep_cnt].bEndpointAddress);
	    dbg("Endpoint %d: Attr = %d", ep_cnt, endpoint[ep_cnt].bmAttributes);
	    dbg("Endpoint %d: MPS = %d", ep_cnt, endpoint[ep_cnt].wMaxPacketSize);
	    dbg("Endpoint %d: Interval = %d", ep_cnt, endpoint[ep_cnt].bInterval);
	    dbg("Endpoint %d: Refresh = %d", ep_cnt, endpoint[ep_cnt].bRefresh);
	    dbg("Endpoint %d: SyncAddr = %d", ep_cnt, endpoint[ep_cnt].bSynchAddress);
	
	    if (!have_bulk_out && IS_EP_BULK_OUT(endpoint[ep_cnt])) {
	        ep_cnt++;
	        have_bulk_out = ep_cnt;
	        continue;
	    }
	    if (!have_response_intr && IS_EP_INTR(endpoint[ep_cnt]) && endpoint[ep_cnt].wMaxPacketSize == 32) {
	        ep_cnt++;
	        have_response_intr = ep_cnt;
	        continue;
	    }
	    if (!have_event_intr && IS_EP_INTR(endpoint[ep_cnt]) && endpoint[ep_cnt].wMaxPacketSize == 8) {
	        ep_cnt++;
	        have_event_intr = ep_cnt;
	        continue;
	    }
	    
        dbg("probe_reader: Undetected endpoint.");
	    return NULL;	/* Shouldn't ever get here unless we have something weird */
    }
    
    if (!have_bulk_out || !have_response_intr || !have_event_intr) {
    	dbg("probe_reader: Two bulk endpoints and one interrupt endpoint required.");
	    return NULL;
    }
    
	down(&rdr_mutex);

    /*
     Determine a minor number and initialize the structure associated with it.  
     The problem with this is that we are counting on the fact that the user will 
     sequentially add device nodes for the reader devices. 
    */
	for (minor = 0; minor < READER_MAX_MINOR; minor++) {
		if (!reader_data_table[minor])
			break;
	}

    /* Check to make sure that the last slot isn't already taken */
	if (reader_data_table[minor]) {
		err("probe_reader: No more minor devices remaining.");
        up(&rdr_mutex);
		return NULL;
	}

	dbg("probe_reader: Allocated minor:%d", minor);

    /* allocate memory for a reader_data structure */
    
    if (!(data = kmalloc (sizeof (struct reader_data), GFP_KERNEL))) {
	    err("probe_reader: Out of memory.");
        up(&rdr_mutex);
	    return NULL;
    }
    memset (data, 0, sizeof(struct reader_data));

    init_MUTEX(&(data->gen_lock)); /* initialized to be unlocked */
    
    /* Setup a handler for the event interrupt */
    if (have_event_intr) {
	    FILL_INT_URB(&data->event_irq, dev,
		         usb_rcvintpipe(dev, have_event_intr),
		         &data->eventBuffer, EVENT_BUFFER_SIZE, event_irq_callback, data, 
	             endpoint[(int)have_event_intr - 1].bInterval);
	    if (usb_submit_urb(&data->event_irq)) {
	        err("probe_reader: Unable to allocate INT URB.");
            kfree(data);
            up(&rdr_mutex);
            return NULL;
        }
    }

    /* Setup a handler for the data interrupt */
    if (have_response_intr) {
	    FILL_INT_URB(&data->response_irq, dev,
		         usb_rcvintpipe(dev, have_response_intr),
		         &data->responseBuffer, RESPONSE_BUFFER_SIZE, response_irq_callback, data, 
	             endpoint[(int)have_response_intr - 1].bInterval);
	    if (usb_submit_urb(&data->response_irq)) {
	        err("probe_reader: Unable to allocate INT URB.");
            kfree(data);
            up(&rdr_mutex);
            return NULL;
        }
    }
    
	sprintf(name, "AseIIIReader%d", minor);
	
	data->devfs = devfs_register(usb_devfs_handle, name,
				    DEVFS_FL_DEFAULT, READER_MAJOR,
				    READER_BASE_MNR + minor,
				    S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP |
				    S_IWGRP | S_IROTH | S_IWOTH, &usb_reader_fops, NULL);
	if (data->devfs == NULL)
		dbg("AseIIIReader%d: device node registration failed", minor);

    reader_data_table[minor] = data;

    up(&rdr_mutex);

    /* Ok, now initialize all the relevant values  */
    data->bulk_out_ep = have_bulk_out;
    data->response_intr_ep = have_response_intr;
    data->event_intr_ep = have_event_intr;
    data->present = 1;
    data->reader_dev = dev;
    data->isopen = 0;
    data->minor = minor;
    
    data->rwlockEvent = RW_LOCK_UNLOCKED;
    data->rwlockResponse = RW_LOCK_UNLOCKED;

    data->curResponsePos = data->lastResponsePos = data->curEventPos = data->lastEventPos = 0;

    return data;
}



/*****************************************************************************
*
*****************************************************************************/
static void disconnect_reader(struct usb_device *dev, void *ptr)
{
    struct reader_data *data = (struct reader_data *) ptr;

    down(&rdr_mutex);
    down(&(data->gen_lock));

    /*
     if the reader has been removed WITHOUT calling the close() function first, then
     there may be some URB requests pending or that we are in the middle of read/write
     to/from the driver. this means we cannot release the driver's resources here - we
     only set the present flag to false, and the next time close() will be called it
     will do the cleanup code of that should have been done here.
    */
    if (data->isopen) {
        data->present = 0;
        up(&(data->gen_lock));
        up(&rdr_mutex);
        return;
    }

    /* the reader is closed, so its resources can be released */
    if(data->response_intr_ep) {
	    usb_unlink_urb(&data->response_irq); 
    }
    if(data->event_intr_ep) {
	    usb_unlink_urb(&data->event_irq); 
    }

    usb_driver_release_interface(&reader_driver, 
                                 &data->reader_dev->actconfig->interface[data->ifnum]);

    devfs_unregister(data->devfs);
    reader_data_table[data->minor] = NULL;
    up(&(data->gen_lock));
    kfree (data);
    up(&rdr_mutex);
}



static struct usb_driver reader_driver = {
    name:	    "ASE III USB Smart Card Reader",
    probe:	    probe_reader,
    disconnect: disconnect_reader,
    fops:	    &usb_reader_fops,
    minor:		READER_BASE_MNR, 
    id_table:	reader_device_ids, 
};





/*****************************************************************************
*
*****************************************************************************/
/*
 * Functions called at module load and unload time: only register and
 * unregister the USB callbacks and the misc entry point
 */
int init_module(void)
{
    int retval = usb_register(&reader_driver);
    if (retval < 0) {
        dbg("ASE III USB reader registeration returned %d. Exiting...", retval);
        return -1;
    }

    info("ASE III USB reader registered.");

    return 0;
}

void cleanup_module(void)
{
    usb_deregister(&reader_driver);

    info("ASE III USB reader deregistered.");

}



