#include "Ase.h"


/* ASE reader generic command types */


/*****************************************************************************
*
*****************************************************************************/
/* returns 0 on success, negative number otherwise */
int checkValidity (int retVal, int len, int actual, char* message) {
    if (retVal < 0 || actual != len) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "%s   Error = %d\n", message, retVal);
#endif
        return retVal; 
    }
    return ASE_OK;
}

/*****************************************************************************
*
*****************************************************************************/
static void cleanResponseBuffer (reader* globalData) {
    int retVal, actual;
    char dummy[EVENT_BUFFER_SIZE];

    retVal = ClearKernelResponseBuffer(globalData);

/*
    do {
        retVal = readCloseResponse(globalData, 0, 1, dummy, &actual);
    } while (actual != 0 && retVal >= 0);
*/
}


/*****************************************************************************
*
*****************************************************************************/
/* returns 0 on success, negative number otherwise
 outBuf should be large enough to contain the close response */
int sendCloseResponseCommand (reader* globalData, char* cmd, int len, 
                              char* outBuf, int* outBufLen) {
    int retVal, actual, i;
    uchar oneByte, cs;

    globalData->curPos = globalData->lastPos = 0;

    /* send the command */

    retVal = writeToReader(globalData, cmd, len, &actual);
    if (checkValidity(retVal, len, actual, "sendCloseResponseCommand - Error! in command write.\n")) {
        cleanResponseBuffer(globalData);
        return retVal; 
    }

    /* read the close response */

    /* read packet header and verify it's ok */
    retVal = readCloseResponse(globalData, 0, 1, &oneByte, &actual);
    if (checkValidity(retVal, 1, actual, "sendCloseResponseCommand - Error! in packet header read.\n")) {
        cleanResponseBuffer(globalData);
        return retVal; 
    }

    cs = oneByte;

    if (oneByte != 0x10) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "sendCloseResponseCommand - Error! packet header is 0x%x%x.\n", (oneByte & 0xF0) >> 4, oneByte & 0x0F);
#endif
        cleanResponseBuffer(globalData);
        return ackStatus(oneByte); 
    }

    /* read data length */
    retVal = readCloseResponse(globalData, 0, 1, &oneByte, &actual);
    if (checkValidity(retVal, 1, actual, "sendCloseResponseCommand - Error! in data length read.\n")) {
        cleanResponseBuffer(globalData);
        return retVal; 
    }

    cs ^= oneByte;

    /* read data */
    retVal = readCloseResponse(globalData, 0, oneByte, outBuf, outBufLen);
    if (checkValidity(retVal, oneByte, *outBufLen, "sendCloseResponseCommand - Error! in data read.\n")) {
        cleanResponseBuffer(globalData);
        return retVal; 
    }

    for (i = 0 ; i < *outBufLen ; ++i)
        cs ^= outBuf[i];

    /* read the checksum */
    retVal = readCloseResponse(globalData, 0, 1, &oneByte, &actual);
    if (checkValidity(retVal, 1, actual, "sendCloseResponseCommand - Error! in checksum read.\n")) {
        cleanResponseBuffer(globalData);
        return retVal; 
    }

    if (cs != oneByte) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "sendCloseResponseCommand - Error! invalid checksum.\n");
#endif
        cleanResponseBuffer(globalData);
        return ASE_ERROR_CHECKSUM; 
    }

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
/* returns 0 on success, negative number otherwise
   outBuf will contain the ACK/NAK byte sent as response */
int sendControlCommand (reader* globalData, char* cmd, int len, 
                        char* outBuf, int* outBufLen) {
    int retVal, actual, i;
    uchar oneByte, cs;

    globalData->curPos = globalData->lastPos = 0;

    /* send the command */

    retVal = writeToReader(globalData, cmd, len, &actual);
    if (checkValidity(retVal, len, actual, "sendControlCommand - Error! in command write.\n")) {
        cleanResponseBuffer(globalData);
        return retVal; 
    }

    /* read the Ack control byte */

    retVal = readCloseResponse(globalData, 0, 1, outBuf, outBufLen);
    if (checkValidity(retVal, 1, *outBufLen, "sendControlCommand - Error! in ack read.\n")) {
    	cleanResponseBuffer(globalData);
        return retVal; 
    }

    if (outBuf[0] != 0x20) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "sendControlCommand - Error! ack is 0x%x%x.\n", (outBuf[0] & 0xF0) >> 4, outBuf[0] & 0x0F);
#endif
        cleanResponseBuffer(globalData);
        return ackStatus(outBuf[0]); 
    }

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
/* returns 0 on success, negative number otherwise */
static int finishOpenResponseCommand (reader* globalData, char socket, int errorStatus, uchar cs, uchar checksum) {

    int retVal, actual;
    char dummy, stopByte;

    if (errorStatus == ASE_OK) {
        /* send STOP byte */
        stopByte = 0x60;
        retVal = writeToReader(globalData, &stopByte, 1, &actual);
        if (checkValidity(retVal, 1, actual, "finishOpenResponseCommand: Error! could not write STOP byte.\n")) {
            cleanResponseBuffer(globalData);
            return retVal; 
        }

        /* read checksum + CEEI */
        retVal = readCloseResponse(globalData, 0, 1, &checksum, &actual);
        if (checkValidity(retVal, 1, actual, "finishOpenResponseCommand: Error! could not read checksum.\n")) {
            cleanResponseBuffer(globalData);
            return retVal; 
        }

        if (checksum & 0x01) {
            (globalData->cards[socket].ceei)--;
        }

        /* verify the checksum */
        cs &= 0xFE;
        checksum &= 0xFE;
        if (cs != checksum) {
#ifdef ASE_DEBUG
            syslog(LOG_INFO, "finishOpenResponseCommand: Error! checksum = 0x%x%x cs = 0x%x%x.\n", PRINT_CHAR(checksum), PRINT_CHAR(cs));
#endif
            cleanResponseBuffer(globalData);
            return ASE_ERROR_CHECKSUM; 
        }
    }
    else { /* errorStatus != ASE_OK */
        /* send STOP byte */
        stopByte = 0x60;
        retVal = writeToReader(globalData, &stopByte, 1, &actual);
        if (checkValidity(retVal, 1, actual, "finishOpenResponseCommand: Error! could not write STOP byte.\n")) {
            cleanResponseBuffer(globalData);
            return retVal; 
        }

	    /* read the checksum */
        retVal = readCloseResponse(globalData, 0, 1, &checksum, &actual);

        /* clean the repsonse data */
        cleanResponseBuffer(globalData);
        globalData->curPos = globalData->lastPos = 0;

        /* send a second STOP byte */
        stopByte = 0x60;
        retVal = writeToReader(globalData, &stopByte, 1, &actual);
        if (checkValidity(retVal, 1, actual, "finishOpenResponseCommand: Error! could not write STOP byte.\n")) 
            return retVal; 

        /* read checksum that should be 0x00 */
        retVal = readCloseResponse(globalData, 0, 1, &checksum, &actual);
        if (checkValidity(retVal, 1, actual, "finishOpenResponseCommand: Error! could not read checksum.\n")) 
            return retVal; 

        if (checksum != 0x00) { 
#ifdef ASE_DEBUG
            syslog(LOG_INFO, "finishOpenResponseCommand: Error! checksum = 0x%x%x and not 0x00\n", PRINT_CHAR(checksum));
#endif
            cleanResponseBuffer(globalData);
            return ASE_ERROR_CHECKSUM; 
        }
    }


    return ASE_OK;
}

/*****************************************************************************
*
*****************************************************************************/
/* returns 0 on success, negative number otherwise
 outBuf will contain the ACK/NAK byte sent as response
 The arg is used for transfering data to openResponseProccessor() function */
int sendOpenResponseCommand (reader* globalData, char socket, char* cmd, int len,
                             char* outBuf, int* outBufLen, 
                             openResponseProccessor orp, unsigned long arg,
                             int sendStopByte) {
    int retVal, actual, i;
    uchar oneByte, cs, checksum, stopByte;

    globalData->curPos = globalData->lastPos = 0;

    /* send the command */
    if (len) { /* because T=1 blocks may contain 0 bytes of data, but still we need to read the response */
	retVal = writeToReader(globalData, cmd, len, &actual);
	if (checkValidity(retVal, len, actual, "sendOpenResponseCommand - Error! in command write.\n")) {
	    finishOpenResponseCommand(globalData, socket, retVal, cs, checksum);
	    return retVal; 
	}
    }

    /* read the open response */

    /* read packet header and verify it's ok */
    retVal = readOpenResponse(globalData, socket, 1, &oneByte, &actual, 1);
    if (checkValidity(retVal, 1, actual, "sendOpenResponseCommand - Error! in packet header read.\n")) {
        finishOpenResponseCommand(globalData, socket, retVal, cs, checksum);
        return retVal; 
    }

    cs = oneByte;

    if ((oneByte & 0xF0) != 0x70) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "sendOpenResponseCommand - Error! packet header is 0x%x%x.\n", (oneByte & 0xF0) >> 4, oneByte & 0x0F);
#endif
        finishOpenResponseCommand(globalData, socket, ackStatus(oneByte), cs, checksum);
        return ackStatus(oneByte); 
    }

    if (oneByte & 0x01)
        (globalData->cards[socket].ceei)--;

    /* read the data (cs is computed in the function on the response bytes!!!) */
    retVal = (*orp)(globalData, socket, arg, &cs, outBuf, outBufLen);
    if (retVal < 0) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "sendOpenResponseCommand - Error! in response read.\n");
#endif
        finishOpenResponseCommand(globalData, socket, retVal, cs, checksum);
        return retVal; 
    }

    if (sendStopByte)
        retVal = finishOpenResponseCommand(globalData, socket, ASE_OK, cs, checksum);

    return retVal; 
}




