#include "Ase.h"



/*****************************************************************************
*
*****************************************************************************/
#define IOCTL_RESPONSE_BUFFER_SIZE    512
#define IOCTL_EVENT_BUFFER_SIZE       16

typedef struct {
    int     len;
    int     timeout; 
    char    data[IOCTL_RESPONSE_BUFFER_SIZE];
} ioctlResponseStruct;


typedef struct {
    int     timeout; 
    char    data[IOCTL_EVENT_BUFFER_SIZE];
} ioctlEventStruct;


typedef struct {
    char    present;
} ioctlStatusStruct;


typedef struct {
    char    clear; // redundant
} ioctlClearStruct;


/* some ioctl - make sure they are unregistered!!! */
#define IOCTL_ASE_RESPONSE      _IOWR('U', 0x20, ioctlResponseStruct)
#define IOCTL_ASE_EVENT         _IOWR('U', 0x21, ioctlEventStruct)
#define IOCTL_ASE_STATUS        _IOWR('U', 0x22, ioctlStatusStruct)
#define IOCTL_ASE_CLEAR_BUFFER  _IOWR('U', 0x23, ioctlClearStruct)


#define OPEN_RESPONSE_TIMEOUT       100 /* num of retries and NOT ms */
#define CLOSE_RESPONSE_TIMEOUT      10  /* num of retries and NOT ms */

#define OPEN_RESPONSE_CHUNK_SIZE     512 /*128 - MUST be a multiple of 32!!! */
#define CLOSE_RESPONSE_CHUNK_SIZE    32  /* num of bytes to try to read in ioctl */
#define EVENT_CHUNK_SIZE             8   /* num of bytes to try to read in ioctl */

#define EID_MASK    0xC0
#define EID(c)      ((c & EID_MASK) >> 6)


/*****************************************************************************
*
*****************************************************************************/
/* reader status */
#define STS_PID_OK                  0x00
#define STS_PID_ERROR               0x01
#define STS_CNT_ERROR               0x02
#define STS_TRUNC_ERROR             0x03
#define STS_LEN_ERROR               0x04
#define STS_UNKNOWN_CMD_ERROR       0x05
#define STS_TIMEOUT_ERROR           0x06
#define STS_CS_ERROR                0x07
#define STS_INVALID_PARAM_ERROR     0x08
#define STS_CMD_FAILED_ERROR        0x09
#define STS_NO_CARD_ERROR           0x0A
#define STS_CARD_NOT_POWERED_ERROR  0x0B
#define STS_COMM_ERROR              0x0C


/* Card events (EID == 0) */
#define EID_CARD1_NOT_PRESENT       0x00
#define EID_CARD1_NOT_PRESENT       0x10
#define EID_CARD1_INSERTED          0x01
#define EID_CARD2_INSERTED          0x11
/* Character error events (EID == 1) */
#define EID_PARITY_ERROR            0x41
#define EID_FIFO_OVERRUN            0x42
#define EID_FRAMING_ERROR           0x43
#define EID_EARLY_ANSWER            0x44
#define EID_TIMEOUT                 0x45
/* Synchronization events (EID == 2) */
#define EID_CARD1_POWERED_ON        0x81
#define EID_CARD2_POWERED_ON        0x91
#define EID_CARD1_POWERED_OFF       0x80
#define EID_CARD2_POWERED_OFF       0x90
/* General events (EID == 3) */
#define EID_OVERHEATING             0xC1



/*****************************************************************************
*
*****************************************************************************/
static int readEvents (reader* globalData, char* outBuf, int* outBufLen);

/*****************************************************************************
*
*****************************************************************************/
/* returns 0 if no error, status code otherwise */
int ackStatus (uchar ackByte) {
    int retVal;

    ackByte &= 0x0F;

    switch (ackByte) {
        case STS_PID_ERROR:
            retVal = ASE_READER_PID_ERROR;
            break;
        case STS_CNT_ERROR:
            retVal = ASE_READER_CNT_ERROR;
            break;
        case STS_TRUNC_ERROR:
            retVal = ASE_READER_TRUNC_ERROR;
            break;
        case STS_LEN_ERROR:
            retVal = ASE_READER_LEN_ERROR;
            break;
        case STS_UNKNOWN_CMD_ERROR:
            retVal = ASE_READER_UNKNOWN_CMD_ERROR;
            break;
        case STS_TIMEOUT_ERROR:
            retVal = ASE_READER_TIMEOUT_ERROR;
            break;
        case STS_CS_ERROR:
            retVal = ASE_READER_CS_ERROR;
            break;
        case STS_INVALID_PARAM_ERROR:
            retVal = ASE_READER_INVALID_PARAM_ERROR;
            break;
        case STS_CMD_FAILED_ERROR:
            retVal = ASE_READER_CMD_FAILED_ERROR;
            break;
        case STS_NO_CARD_ERROR:
            retVal = ASE_READER_NO_CARD_ERROR;
            break;
        case STS_CARD_NOT_POWERED_ERROR:
            retVal = ASE_READER_CARD_NOT_POWERED_ERROR;
            break;
        case STS_COMM_ERROR:
            retVal = ASE_READER_COMM_ERROR;
            break;
        default:
            retVal = 0;
            break;
    }

#ifdef ASE_DEBUG
    if (retVal) 
        syslog(LOG_INFO, "ackStatus: Error! status = %d\n", ackByte);
#endif

    return retVal;
}

/*****************************************************************************
*
*****************************************************************************/
/* returns 0 if not an error, error code otherwise.
 change global data status if neccesary */
int parseEvent (reader* globalData, char socket, uchar event) {
    int retVal;
    uchar eid = EID(event), skn;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "=============> Event = 0x%x%x <===============\n", PRINT_CHAR(event));
#endif

    switch (eid) {
        case 0: /* card */
	        skn = (event & 0x30) >> 4;
	        globalData->cards[skn].status = event & 0x01;
	        if (!(event & 0x01) && (skn == socket))
	            return ASE_READER_NO_CARD_ERROR;
	        break;
        case 1: /* character */
	        switch (event & 0x07) {
	            case 1:
	                return ASE_ERROR_PARITY;
	            case 2:
	                return ASE_ERROR_FIFO_OVERRUN;
	            case 3:
	                return ASE_ERROR_FRAMING;
	            case 4:
	                return ASE_ERROR_EARLY_ANSWER;
                case 5: {
      	                globalData->cards[(event & 0x30) >> 4].ceei++; 
                        if (((event & 0x30) >> 4) == socket)
                            return ASE_ERROR_TIMEOUT;
	                }
            }
	        break;
        case 2: /* sync */
	        skn = (event & 0x30) >> 4;
	        if (event & 0x01)
	            globalData->cards[skn].status = 2;
	        else
	            globalData->cards[skn].status = 1;
	        /*
	        if (!(event & 0x01))
	            return ASE_READER_CARD_NOT_POWERED_ERROR;
	        */
	        return ASE_OK;
	        break;
        case 3: /* global */
	        if (event & 0x07 == 1) { /* overheating */
	            globalData->readerStarted = 0; 
	            return ASE_ERROR_OVERHEATING;
	        }
	        break;
    }

    return ASE_OK;
}


/*****************************************************************************
*
*****************************************************************************/
int DeviceIsConnected (reader* globalData) {
    ioctlStatusStruct resp;
    int retVal;

    retVal = ioctl(globalData->fd, IOCTL_ASE_STATUS, &resp);

    if (retVal == ASE_OK) { 
        return resp.present;
    }
    else {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "DeviceIsConnected - Error! actual = %d\n", retVal);
#endif
        /* we assume there is no device if there is an error ???????? */
        return 0;
    }
}


/*****************************************************************************
*
*****************************************************************************/
int ClearKernelResponseBuffer (reader* globalData) {
    ioctlClearStruct resp;
    int retVal;

    retVal = ioctl(globalData->fd, IOCTL_ASE_CLEAR_BUFFER, &resp);

    if (retVal == ASE_OK) { 
        return retVal;
    }
    else {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "ClearKernelResponseBuffer - Error! retVal = %d\n", retVal);
#endif
        return 0;
    }
}



/*****************************************************************************
*
*****************************************************************************/
int writeToReader (reader* globalData, char* data, int len, int* actual) {
    int i;

    *actual = write(globalData->fd, data, len);

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "writeToReader: wrote %d bytes. data = ", *actual);
    for (i = 0 ; i < *actual ; ++i)
	    syslog(LOG_INFO, "0x%x%x ", PRINT_CHAR(data[i]));
    syslog(LOG_INFO, "\n");
#endif

    if (*actual != len) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "writeToReader - Error! len = %d, actual = %d\n", len, *actual);
#endif
        return ASE_ERROR_WRITE;
    }

    return ASE_OK;
}


/*****************************************************************************
*
*****************************************************************************/
/* calls the ioctl, and stores the data read in the global buffer
 returns: negative - error, 0 - success. actual returns the number of bytes read */
static int pollCloseResponseFromReader (reader* globalData, int len, int timeout, int* actual) {
    int i;
    ioctlResponseStruct resp;

    resp.len = len;
    resp.timeout = timeout;

    *actual = ioctl(globalData->fd, IOCTL_ASE_RESPONSE, &resp);

    if (*actual > 0) { /* copy data to buffer */
        for (i = 0 ; i < *actual ; ++i) {
            globalData->response[globalData->lastPos++] = resp.data[i];
            globalData->lastPos %= RESPONSE_BUFFER_SIZE;
        }
    }
    else if (*actual < 0) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "pollCloseResponseFromReader - Error! actual = %d\n", *actual);
#endif
        return ASE_ERROR_IOCTL_RESPONSE;
    }

    return ASE_OK;
}

/*****************************************************************************
*
*****************************************************************************/
/* calls the ioctl, and stores the data read in the global buffer
 cleans all the USB padding before copying the data to the buffer
 returns: negative - error, 0 - success. actual returns the number of bytes read */
static int pollOpenResponseFromReader (reader* globalData, int len, int timeout, int* actual) {
    int i, j;
    char relevant;
    ioctlResponseStruct resp;

    resp.len = OPEN_RESPONSE_CHUNK_SIZE;
    resp.timeout = timeout;

    *actual = ioctl(globalData->fd, IOCTL_ASE_RESPONSE, &resp);

#ifdef ASE_DEBUG
    /*        syslog(LOG_INFO, "pollOpenResponseFromReader - actual = %d\n", *actual);*/
#endif

    if (*actual > 0) { /* copy data to buffer */
        for (i = 0 ; i < *actual ; i += 32) {
            /* first byte is the number of valid bytes in current chunk */
            relevant = resp.data[i];
            if (!relevant) /* no valid bytes */
                continue;

            for (j = 0 ; j < relevant ; j++) {
                globalData->response[globalData->lastPos++] = resp.data[i + j + 1];
                globalData->lastPos %= RESPONSE_BUFFER_SIZE;
            }
        }
    }
    else if (*actual < 0)  {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "pollOpenResponseFromReader - Error! actual = %d\n", *actual);
#endif
        return ASE_ERROR_IOCTL_RESPONSE;
    }

    return ASE_OK;
}





/*****************************************************************************
*
*****************************************************************************/
/* returns 0 on success; negative value otherwise */
int readCloseResponse (reader* globalData, char socket, int num, char* outBuf, int* outBufLen) {
    int retVal, retryTimes, len, remain = num, stopLoop = 0;
    int i = 0;
    char events[EVENT_BUFFER_SIZE];
    int eventsLen = 0, eventError = ASE_OK;

    *outBufLen = 0;

    retryTimes = 200;
    while (remain > 0 && retryTimes && !stopLoop) {
        /* get data from kernel driver */
        retVal= pollCloseResponseFromReader(globalData, remain, CLOSE_RESPONSE_TIMEOUT, &len);
        if (retVal < 0) {
#ifdef ASE_DEBUG
            syslog(LOG_INFO, "readCloseResponse - Error! retVal = %d\n", retVal);
#endif
            return retVal; 
        }

        /* if something was read, copy it to the outgoing buffer (but not more than requested!) */
        while (remain > 0 && globalData->curPos != globalData->lastPos) { 
            /* there is some data that hasnt been read yet */
            outBuf[*outBufLen] = globalData->response[globalData->curPos];

            (*outBufLen)++;
            remain--;
            globalData->curPos++;
            globalData->curPos %= RESPONSE_BUFFER_SIZE;
        }

        /* if not all the data requested has arrived - check the events */
	    if (*outBufLen != num) {
	        retVal = readEvents(globalData, events, &eventsLen);
	        if (retVal < 0) {
#ifdef ASE_DEBUG
    		    syslog(LOG_INFO, "readCloseResponse - Error! Couldnt get events. retVal = %d\n", retVal);
#endif
	    	    return retVal;
	        }
	        
	        while (eventsLen && !stopLoop) {
		        eventError = parseEvent(globalData, socket, events[eventsLen - 1]);
                if (eventError && eventError != ASE_ERROR_TIMEOUT) {
#ifdef ASE_DEBUG
    		        syslog(LOG_INFO, "readCloseResponse: eventError = %d\n", eventError);
#endif
                    stopLoop = 1;
                }
    		    eventsLen--;
	        }
	        
	        /* parse the events (to update global data) but ignore them here */
	        while (eventsLen > 0) {
	    	    parseEvent(globalData, socket, events[eventsLen - 1]);
		        eventsLen--;
	        }
	        
	        usleep(1000); /* 1 ms */
	        retryTimes--;
	    } /* if (*outBufLen != num) */
	    
	    eventError = ASE_OK;
	
    } /* while (remain > 0 && retryTimes && !stopLoop) */
    
    
#ifdef ASE_DEBUG
    syslog(LOG_INFO, "***readCloseResponse: got %d bytes. data = ", *outBufLen);
    for (i = 0 ; i < *outBufLen ; ++i)
	    syslog(LOG_INFO, "0x%x%x ", PRINT_CHAR(outBuf[i]));
    syslog(LOG_INFO, "\n");
#endif

    
    if (num == *outBufLen) { /* all data arrived */
#ifdef ASE_DEBUG
	    if (eventError != ASE_OK) 
	        syslog(LOG_INFO, "readCloseResponse: Event has been read but all data arrived. eventError = %d\n", eventError);
#endif
        return ASE_OK;
    }
    else { /* part of the data is missing */
        if (!eventError) { /* no event arrived or the parser returned ASE_OK */
#ifdef ASE_DEBUG
	        syslog(LOG_INFO, "readCloseResponse: No error events read when NOT all data arrived. Got %d out of %d\n", *outBufLen, num);
#endif
            return ASE_ERROR_DATA_INCOMPLETE;
    	}
        else {
#ifdef ASE_DEBUG
	        syslog(LOG_INFO, "readCloseResponse: Event has been read when NOT all data arrived. Got %d out of %d. eventError = %d\n", *outBufLen, num, eventError);
#endif
            return eventError;
        }
    }
    
    return ASE_OK;
}


/*****************************************************************************
*
*****************************************************************************/
/* returns 0 on success; negative value otherwise */
int readOpenResponse (reader* globalData, char socket, int num, char* outBuf, int* outBufLen, int startOfResponse) {
    int retVal, retryTimes, len, stopLoop = 0, remain = num;
    int i = 0;
    char events[EVENT_BUFFER_SIZE];
    int eventsLen = 0, eventError = ASE_OK;
    /* wait at least 1ms for each character */
    long cwt = (globalData->cards[socket].cwt > 0 ? globalData->cards[socket].cwt : 1000);
    long chunkNum;

    /* if this is the first character of the card's response, we should wait at least BWT.
     if the active protocol is T=0, then bwt=cwt. if T=1, then bwt may be bigger than cwt. */
    if (startOfResponse)
        cwt = MAX(cwt, globalData->cards[socket].bwt);

    cwt += 1000000; 

    /*
    chunkNum = cwt / 50000;
#ifdef ASE_DEBUG
    syslog(LOG_INFO, "readOpenResponse - chunkNum = %d\n", chunkNum);
#endif
    */

    *outBufLen = 0;

    while (remain && !stopLoop) {

        retryTimes = 32;

        while (remain > 0 && /*chunkNum*/ retryTimes && !stopLoop) {
        
            usleep(cwt >> 5); //50000); 

            /* get data from kernel driver */
            retVal= pollOpenResponseFromReader(globalData, remain, OPEN_RESPONSE_TIMEOUT, &len);
            if (retVal < 0) {
#ifdef ASE_DEBUG
                syslog(LOG_INFO, "readOpenResponse - Error! retVal = %d\n", retVal);
#endif
                return retVal;
            }

            /* if something was read, copy it to the outgoing buffer (but not more than requested!) */
            while (remain > 0 && globalData->curPos != globalData->lastPos) { 
                /* there is some data that hasnt been read yet */
                outBuf[*outBufLen] = globalData->response[globalData->curPos];

                (*outBufLen)++;
                remain--;
                globalData->curPos++;
                globalData->curPos %= RESPONSE_BUFFER_SIZE;
            }

            /* if not all the data requested has arrived - check the events */
            if (*outBufLen != num) {
                retVal = readEvents(globalData, events, &eventsLen);
                if (retVal < 0) {
#ifdef ASE_DEBUG
                    syslog(LOG_INFO, "readOpenResponse - Error! Couldnt get events. retVal = %d\n", retVal);
#endif
                    return retVal;
                }
                /* check all the events read */
                while (eventsLen && !stopLoop) {
                    eventError = parseEvent(globalData, socket, events[eventsLen - 1]);
                    /* There is an error only if the timeout event belongs to this command (ceei>0) or another error event */
                    if (eventError) {
                        if (eventError == ASE_ERROR_TIMEOUT && globalData->cards[socket].ceei <= 0) {
                           continue;
                        }
                        else {
#ifdef ASE_DEBUG
                            syslog(LOG_INFO, "readOpenResponse: eventError = %d retryTimes = %d\n", eventError, retryTimes);
#endif
                            stopLoop = 1;
                        }
                    }
                    eventsLen--;
                }
                /* if any event is left, at least update global data if neccesary */
                while (eventsLen > 0) {
                    parseEvent(globalData, socket, events[eventsLen - 1]);
                    eventsLen--;
                }

                retryTimes--;//chunkNum--;
            }

            eventError = ASE_OK;
        }
    }
    
    
#ifdef ASE_DEBUG
    syslog(LOG_INFO, "***readOpenResponse: got %d bytes. data = ", *outBufLen);
    for (i = 0 ; i < *outBufLen ; ++i)
	    syslog(LOG_INFO, "0x%x%x ", PRINT_CHAR(outBuf[i]));
    syslog(LOG_INFO, "\n");
#endif

    if (num == *outBufLen) { /* all data arrived */
    	if (eventError != ASE_OK) 
#ifdef ASE_DEBUG
	        syslog(LOG_INFO, "readOpenResponse: Event has been read but all data arrived. eventError = %d\n", eventError);
#endif
        return ASE_OK;
    }
    else { /* part of the data is missing */
        if (!eventError) { /* no event arrived or the parser returned ASE_OK */
#ifdef ASE_DEBUG
    	    syslog(LOG_INFO, "readOpenResponse: No error events read when NOT all data arrived. Got %d out of %d\n", *outBufLen, num);
#endif
            return ASE_ERROR_DATA_INCOMPLETE;
    	}
        else {
#ifdef ASE_DEBUG
	        syslog(LOG_INFO, "readOpenResponse: Event has been read when NOT all data arrived. Got %d out of %d. eventError = %d\n", *outBufLen, num, eventError);
#endif
            return eventError;
        }
    }

    return ASE_OK;
}



/*****************************************************************************
*
*****************************************************************************/
/* calls the ioctl, and stores the data read in the global buffer
 returns: negative - error, 0 - success. actual returns the number of bytes read */
static int pollEventFromReader (reader* globalData) {
    int i, actual;
    ioctlEventStruct resp;

    resp.timeout = 5;

    actual = ioctl(globalData->fd, IOCTL_ASE_EVENT, &resp);

    if (actual > 0) { /* copy data to buffer */
        for (i = 0 ; i < actual ; ++i) {
            globalData->events[globalData->lastEvent++] = resp.data[i];
            globalData->lastEvent %= EVENT_BUFFER_SIZE;
        }
    }
    else if (actual < 0)  {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "pollEventFromReader - Error! actual = %d\n", actual);
#endif
        return ASE_ERROR_IOCTL_EVENT;
    }

    return ASE_OK;
}


/*****************************************************************************
*
*****************************************************************************/
/* returns 0 on success; negative value otherwise */
static int readEvents (reader* globalData, char* outBuf, int* outBufLen) {
    int retVal;
    int i = 0;

    *outBufLen = 0;

    retVal= pollEventFromReader(globalData);
    if (retVal < 0) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "readEvent - Error! retVal = %d\n", retVal);
#endif
        return retVal; 
    }

    while (globalData->curEvent != globalData->lastEvent) { 
        /* there is some data that hasnt been read yet */
        outBuf[*outBufLen] = globalData->events[globalData->curEvent];

        (*outBufLen)++;
        globalData->curEvent++;
        globalData->curEvent %= EVENT_BUFFER_SIZE;
    }

    return ASE_OK;
}




